
/*
Main wrapper module
*/

var AppId, AppPass, OCR, OcrHost, Port, fs, http, sys, url, xml2json;

OcrHost = "cloud.ocrsdk.com";

AppId = process.env.ABBYY_APPID || "app-id-here";

AppPass = process.env.ABBYY_PWD || "app-passwd-here";

Port = "443";

http = require("http");

fs = require("fs");

sys = require("util");

xml2json = require("xml2json");

url = require("url");

/*
Class that implement api calls to abbyy cloud api.
*/

OCR = (function() {

  function OCR() {}

  /*
    Create options, that passed to request object.
  
    @param {String} path Relative path, starting from "/"
    @param {"POST"|"GET"|"PUT"} method Http method
    @param {Object} headers Client headers, authorization will be added forcibly.
    @param {String} port Port :default "80"
    @param {String} host Hotsname :default "cloud.ocrsdk.com"
    @return {Object} options Options dict
  */

  OCR.prototype._createOptions = function(path, method, headers, port, host) {
    var opts;
    if (headers == null) headers = {};
    if (port == null) port = "80";
    if (host == null) host = OcrHost;
    opts = {
      host: host,
      port: port,
      path: path,
      method: method,
      headers: headers
    };
    if (opts.headers.Authorization == null) {
      opts.headers.Authorization = "Basic " + new Buffer("" + AppId + ":" + AppPass).toString("base64");
    }
    return opts;
  };

  /*
    Create options, that passed to request object.
    @see `_createOptions`
  
    @param {String} path Full path, starting from "http"
    @param {"POST"|"GET"|"PUT"} method Http method
    @param {Object} headers Client headers, authorization will be added forcibly.
    @return {Object} options Options dict
  */

  OCR.prototype._createOptionsFromUrl = function(fullpath, method, headers) {
    var parsedUrl, port;
    method || (method = "GET");
    parsedUrl = url.parse(fullpath);
    port = parsedUrl.protocol === "https" ? "443" : "80";
    return this._createOptions(parsedUrl.path, method, headers, port, parsedUrl.host);
  };

  /*
    Get binary data from server. This method is buggy.
  
    @param {Object} opts Request options, @see _createOptionsFromUrl
    @param {Function} fn Callback function, that accept 1) error (or null) 2) data.
  */

  OCR.prototype._getServerBinaryAnswer = function(opts, fn) {
    var ind, req, resData;
    resData = new Buffer(10485760);
    ind = 0;
    req = http.request(opts, function(res) {
      res.setEncoding("binary");
      res.on("data", function(chunk) {
        return ind += resData.write(chunk, ind);
      });
      res.on("end", function() {
        return fn(null, resData);
      });
      return res.on("error", function(err) {
        return fn(err);
      });
    });
    return req.end();
  };

  /*
    Post/get data to server and retrieve answer.
    If opts method set to "GET", request will be sended authomatically, othervise, you need
    to forse request after calling this method by applying `res.end()`.
  
    @param {Object} opts Request options, @see _createOptionsFromUrl
    @param {Function} fn Callback function, that accept 1) error (or null) 2) data.
    @return {Object} req Request object, that can be used to post data and perform "post"/"put" requests.
  */

  OCR.prototype._getServerAnswer = function(opts, fn) {
    var req, resData;
    resData = "";
    req = http.request(opts, function(res) {
      res.setEncoding("utf8");
      res.on("data", function(chunk) {
        return resData += chunk;
      });
      res.on("end", function() {
        return fn(null, resData);
      });
      return res.on("error", function(err) {
        return fn(err);
      });
    });
    if ((opts.method != null) && opts.method.toLowerCase() === "get") req.end();
    return req;
  };

  /*
    For post purposes.
  */

  OCR.prototype._generateBoundary = function() {
    return "--------------------------------------------------" + (Date.now()) + "--";
  };

  /*
    Get task status by task id.
  
    @param {String} taskId Task id
    @param {Function} fn Callback function, accept 1) error, 2) task status in json format
  */

  OCR.prototype.getTaskStatus = function(taskId, fn) {
    var getOpts;
    getOpts = this._createOptions("/getTaskStatus?taskId=" + taskId, "get");
    return this._getServerAnswer(getOpts, function(err, data) {
      if (!err) {
        try {
          return fn(null, JSON.parse(xml2json.toJson(data)).response);
        } catch (e) {
          return fn({
            msg: "can't parse status for task " + taskId
          });
        }
      } else {
        return fn({
          msg: "server error"
        });
      }
    });
  };

  /*
    Get list of tasks.
  
    @param {Function} fn Callback function, accept 1) error, 2) list of tasks in json format
  */

  OCR.prototype.listTasks = function(fn) {
    var getOpts;
    getOpts = this._createOptions("/listTasks", "get");
    return this._getServerAnswer(getOpts, function(err, data) {
      if (!err) {
        try {
          return fn(null, JSON.parse(xml2json.toJson(data)).response);
        } catch (e) {
          return fn({
            msg: "can't parse list of tasks"
          });
        }
      } else {
        return fn({
          msg: "server error"
        });
      }
    });
  };

  /*
    Apply ocr to file on local filesystem.
  
    @param {String} filename Absolute/relative path to file
    @param {Object|Function} opts Options, of function, see next param.
          opts.outputFormat - one of "txt", "rtf", "pdfSearchable", "docx", "xml"
          opts.lang  (String, Array of strings) - language, russian, english and maybe others
    @param {Function} fn Callback function, accept 1) error, 2) new task id
  */

  OCR.prototype.applyToFile = function(filename, opts, fn) {
    var buf;
    if ("function" === typeof opts) {
      fn = opts;
      opts = null;
    }
    opts || (opts = {});
    try {
      buf = fs.readFileSync(filename);
      return this.applyToBuffer(buf, opts, fn);
    } catch (e) {
      return fn({
        msg: "can't read file " + filename
      });
    }
  };

  /*
    Apply OCR conversion to buffer, containing image file.
  
    @param {Buffer} buffer Buffer with image
    @param {Object|Function} opts Options, of function, see next param.
          opts.outputFormat - one of "txt", "rtf", "pdfSearchable", "docx", "xml"
          opts.lang  (String, Array of strings) - language, russian, english and maybe others
    @param {Function} fn Callback function, accept 1) error, 2) new task id
  */

  OCR.prototype.applyToBuffer = function(buffer, opts, fn) {
    var boundary, postOpts, postReq, timeout;
    if (opts == null) opts = {};
    opts.outputFormat || (opts.outputFormat = "txt");
    timeout = opts.requestTimeout || 1000;
    opts.lang || (opts.lang = ["russian", "english"]);
    if (opts.lang instanceof Array) opts.lang = opts.lang.join(",");
    boundary = this._generateBoundary();
    postOpts = this._createOptions("/processImage?exportFormat=" + opts.outputFormat + "&language=" + opts.lang, "POST", {
      "Content-Type": "multipart/form-data; boundary= " + boundary,
      "Content-Length": buffer.length
    });
    postReq = this._getServerAnswer(postOpts, function(err, data) {
      var id, msg;
      if (!err) {
        id = data.match(/task id=\"[-a-f\d]+\"/ig);
        if (id) {
          return fn(null, id[0].slice(9, -1));
        } else {
          msg = data.match(/message .*\>.*\<\/message/ig);
          if (msg) {
            return fn({
              msg: msg[0].match(/\>.*\<\//ig)[0].slice(1, -2)
            });
          } else {
            return fn({
              msg: "unknown error"
            });
          }
        }
      } else {
        return fn(err);
      }
    });
    postReq.write(buffer);
    postReq.write(boundary);
    return postReq.end();
  };

  /*
    Get text content after OCR from specified url.
  
    @param {String} srcUrl Url with text content.
    @param {Function} fn Callback function, accept 1) error 2) dictionary with 2 fields: `resultUrl` and `text`
  */

  OCR.prototype.getTextFromUrl = function(srcUrl, fn) {
    var getOpts;
    getOpts = this._createOptionsFromUrl(srcUrl, "GET", false);
    delete getOpts.headers.Authorization;
    return this._getServerBinaryAnswer(getOpts, function(err, data) {
      if (!err) {
        try {
          return fn(null, {
            resultUrl: srcUrl,
            text: data
          });
        } catch (e) {
          return fn({
            msg: "error",
            resultUrl: srcUrl,
            error: e
          });
        }
      } else {
        return fn({
          msg: "error downloading file",
          resultUrl: srcUrl
        });
      }
    });
  };

  /*
    Wait till task end, then get url and content (for text data).
  
    @param {String} taskId Task id
    @param {Object|Function} opts Options, of function, see next param.
          opts.outputFormat - one of "txt", "rtf", "pdfSearchable", "docx", "xml"
          opts.lang  (String, Array of strings) - language, russian, english and maybe others
    @param {Function} fn Callback function, accept 1) error, 2) task object,
                  contain resultUrl and (if opts.outputFormat is set to "txt") text fields.
  */

  OCR.prototype.waitTaskEnd = function(taskId, opts, fn) {
    var downloadInProcess, errors, format, nextTick, timeout;
    var _this = this;
    if ("function" === typeof opts) {
      fn = opts;
      opts = null;
    }
    opts || (opts = {});
    timeout = opts.timeout || 1000;
    format = opts.outputFormat || "txt";
    errors = 0;
    downloadInProcess = false;
    nextTick = function() {
      return _this.getTaskStatus(taskId, function(err, status) {
        var resultUrl, _ref, _ref2, _ref3, _ref4, _ref5, _ref6;
        if (!downloadInProcess) {
          if (!err) {
            if ((status != null ? (_ref = status.task) != null ? (_ref2 = _ref.status) != null ? _ref2.toLowerCase() : void 0 : void 0 : void 0) === "completed") {
              downloadInProcess = true;
              resultUrl = status != null ? (_ref3 = status.task) != null ? _ref3.resultUrl : void 0 : void 0;
              if (resultUrl) {
                if (format === "txt") {
                  return _this.getTextFromUrl(resultUrl, fn);
                } else {
                  return fn(null, {
                    resultUrl: resultUrl
                  });
                }
              }
            } else if ((_ref4 = status != null ? (_ref5 = status.task) != null ? (_ref6 = _ref5.status) != null ? _ref6.toLowerCase() : void 0 : void 0 : void 0) === "queued" || _ref4 === "inprogress") {
              return setTimeout((function() {
                return nextTick();
              }), timeout);
            }
          } else {
            errors++;
            if (errors > 5) {
              return fn({
                msg: "error accessing document"
              });
            } else {
              return setTimeout((function() {
                return nextTick();
              }), timeout);
            }
          }
        }
      });
    };
    return nextTick();
  };

  return OCR;

})();

exports.createWrapper = function() {
  return new OCR();
};
